from __future__ import annotations

import types
import typing

import amulet.level.abc.level

__all__ = [
    "LevelLoaderPathToken",
    "LevelLoaderToken",
    "NoValidLevelLoader",
    "get_level",
]

class LevelLoaderPathToken(LevelLoaderToken):
    def __init__(self, arg0: str) -> None: ...

class LevelLoaderToken:
    @typing.overload
    def __eq__(self, arg0: LevelLoaderToken) -> bool: ...
    @typing.overload
    def __eq__(self, arg0: typing.Any) -> bool | types.NotImplementedType: ...
    def __hash__(self) -> int: ...
    def repr(self) -> str: ...

class NoValidLevelLoader(Exception):
    pass

def get_level(token: LevelLoaderToken) -> amulet.level.abc.level.Level: ...



from __future__ import annotations

import amulet.level.abc.level
import amulet.level.java.dimension
import amulet.level.java.raw_level

__all__ = ["JavaLevel"]

class JavaLevel(
    amulet.level.abc.level.Level,
    amulet.level.abc.level.CompactibleLevel,
    amulet.level.abc.level.DiskLevel,
    amulet.level.abc.level.ReloadableLevel,
):
    @staticmethod
    def create(args: amulet.level.java.raw_level.JavaCreateArgsV1) -> JavaLevel:
        """
        Create a new Java level at the given directory.
        Thread safe.
        """

    @staticmethod
    def load(path: str) -> JavaLevel:
        """
        Load an existing Java level from the given directory.
        Thread safe.
        """

    def get_dimension(
        self, dimension_id: str
    ) -> amulet.level.java.dimension.JavaDimension:
        """
        Get a dimension.
        External Read:SharedReadWrite lock required.
        External ReadWrite:SharedReadWrite lock required when calling code in Dimension (and its children) that need write permission.
        """

    @property
    def raw_level(self) -> amulet.level.java.raw_level.JavaRawLevel:
        """
        Access the raw level instance.
        Before calling any mutating functions, the caller must call :meth:`purge` (optionally saving before)
        External ReadWrite:Unique lock required.
        """



from __future__ import annotations

import amulet.core.chunk
import amulet.utils.lock
import amulet.utils.signal

__all__ = ["ChunkHandle"]

class ChunkHandle:
    def delete_chunk(self) -> None:
        """
        Delete the chunk from the level.
        You must acquire the chunk lock before deleting.
        """

    def exists(self) -> bool:
        """
        Does the chunk exist. This is a quick way to check if the chunk exists without loading it.

        This state may change if the lock is not acquired.

        :return: True if the chunk exists. Calling get on this chunk handle may still throw ChunkLoadError
        """

    def get_chunk(self) -> amulet.core.chunk.Chunk:
        """
        Get a unique copy of the chunk data.

        If you want to edit the chunk, use :meth:`edit` instead.

        If you only want to access/modify parts of the chunk data you can specify the components you want to load.
        This makes it faster because you don't need to load unneeded parts.

        :param components: None to load all components or an iterable of component strings to load.
        :return: A unique copy of the chunk data.
        """

    def set_chunk(self, chunk: amulet.core.chunk.Chunk) -> None:
        """
        Overwrite the chunk data.
        You must acquire the chunk lock before setting.
        If you want to edit the chunk, use :meth:`edit` instead.

        :param chunk: The chunk data to set.
        """

    @property
    def changed(self) -> amulet.utils.signal.Signal[()]:
        """
        Signal emitted when the chunk data changes.
        """

    @property
    def cx(self) -> int:
        """
        The chunk x coordinate.
        """

    @property
    def cz(self) -> int:
        """
        The chunk z coordinate.
        """

    @property
    def dimension_id(self) -> str:
        """
        The dimension identifier this chunk is from.
        """

    @property
    def lock(self) -> amulet.utils.lock.OrderedLock:
        """
        The public lock.
        Thread safe.
        """



from __future__ import annotations

from amulet.level.abc.chunk_handle import ChunkHandle
from amulet.level.abc.dimension import Dimension
from amulet.level.abc.level import Level
from amulet.level.abc.registry import IdRegistry

from . import chunk_handle, dimension, level, registry

__all__ = [
    "ChunkHandle",
    "Dimension",
    "IdRegistry",
    "Level",
    "chunk_handle",
    "dimension",
    "level",
    "registry",
]



from __future__ import annotations

import collections.abc
import types
import typing

import amulet.utils.lock

__all__ = ["IdRegistry"]

class IdRegistry:
    """
    A registry for namespaced ids.
    External synchronisation is required with this class.
    """

    def __contains__(self, arg0: int) -> bool: ...
    def __eq__(self, arg0: typing.Any) -> bool | types.NotImplementedType: ...
    @typing.overload
    def __getitem__(self, index: int) -> tuple[str, str]:
        """
        Convert a numerical id to its namespaced id.
        External shared lock required.
        """

    @typing.overload
    def __getitem__(self, name: tuple[str, str]) -> int:
        """
        Convert a namespaced id to its numerical id.
        External shared lock required.
        """

    def __hash__(self) -> int: ...
    def __init__(self) -> None: ...
    def __iter__(self) -> collections.abc.Iterator[int]:
        """
        An iterable of the numerical ids registered.
        External shared lock required.
        """

    def __len__(self) -> int:
        """
        The number of ids registered.
        External shared lock required.
        """

    def get(
        self, key: int, default: tuple[str, str] | None = None
    ) -> tuple[str, str] | None: ...
    def items(self) -> collections.abc.ItemsView[int, tuple[str, str]]: ...
    def keys(self) -> collections.abc.KeysView[int]: ...
    @typing.overload
    def namespace_id_to_numerical_id(self, name: tuple[str, str]) -> int:
        """
        Convert a namespaced id to its numerical id.
        External shared lock required.
        """

    @typing.overload
    def namespace_id_to_numerical_id(self, namespace: str, base_name: str) -> int:
        """
        Convert a namespaced id to its numerical id.
        External shared lock required.
        """

    def numerical_id_to_namespace_id(self, index: int) -> tuple[str, str]:
        """
        Convert a numerical id to its namespaced id.
        External shared lock required.
        """

    def register_id(self, index: int, name: tuple[str, str]) -> None:
        """
        Convert a namespaced id to its numerical id.
        External unique lock required.
        """

    def values(self) -> collections.abc.ValuesView[tuple[str, str]]: ...
    @property
    def lock(self) -> amulet.utils.lock.SharedLock:
        """
        The public lock.
        Thread safe.
        """



from __future__ import annotations

from builtins import str as JavaInternalDimensionID

from amulet.level.java.chunk_handle import JavaChunkHandle
from amulet.level.java.dimension import JavaDimension
from amulet.level.java.level import JavaLevel
from amulet.level.java.raw_dimension import JavaRawDimension
from amulet.level.java.raw_level import JavaCreateArgsV1, JavaRawLevel

from . import (
    chunk,
    chunk_components,
    chunk_handle,
    dimension,
    level,
    long_array,
    raw_dimension,
    raw_level,
)

__all__ = [
    "JavaChunkHandle",
    "JavaCreateArgsV1",
    "JavaDimension",
    "JavaInternalDimensionID",
    "JavaLevel",
    "JavaRawDimension",
    "JavaRawLevel",
    "chunk",
    "chunk_components",
    "chunk_handle",
    "dimension",
    "level",
    "long_array",
    "raw_dimension",
    "raw_level",
]



from __future__ import annotations

import numpy
import numpy.typing
import typing_extensions
import typing

__all__ = ["decode_long_array", "encode_long_array"]

def decode_long_array(
    long_array: typing.Buffer,
    size: int,
    bits_per_entry: int,
    dense: bool = True,
) -> numpy.ndarray:
    """
    Decode a long array (from BlockStates or Heightmaps)

    :param long_array: Encoded long array
    :param size: The expected size of the returned array
    :param bits_per_entry: The number of bits per entry in the encoded array.
    :param dense: If true the long arrays will be treated as a bit stream. If false they are distinct values with padding
    :return: Decoded array as numpy array
    """

def encode_long_array(
    array: typing.Buffer,
    bits_per_entry: None | int = None,
    dense: bool = True,
    min_bits_per_entry: int = 1,
) -> numpy.typing.NDArray[numpy.uint64]:
    """
    Encode a long array (from BlockStates or Heightmaps)

    :param array: A numpy array of the data to be encoded.
    :param bits_per_entry: The number of bits to use to store each value. If left as None will use the smallest bits per entry.
    :param dense: If true the long arrays will be treated as a bit stream. If false they are distinct values with padding
    :param min_bits_per_entry: The mimimum value that bits_per_entry can be. If it is less than this it will be capped at this value.
    :return: Encoded array as numpy array
    """
